/*
 * adc.c: probe, user config, and user IO (switch and button) reading
 *
 * performs a scanning read on all used ADC inputs and will DMA
 * the results to working variables in one-shot mode. this ADC
 * only uses main voltage as a reference, so scans also include
 * VREFINT for compensation. There is no way to configure a
 * different analog channel (including VREFINT) as the reference
 * voltage - only the analog supply can be the reference.
 * 
 * because most inputs are high impedance, which is not ideal,
 * sampling time is increased to the maximum. this results in
 * a final ADC sample rate of ~33KHz, still well beyond
 * what is needed in this application.
 *
 * file creation: 20231015 0121
 */


#include <stdlib.h>

#include "testo.h"



#define SAMPLE_TIME     LL_ADC_SAMPLINGTIME_239CYCLES_5

#define ADC_CHANNELS	6
#define ADC_HISTLEN		16



static uint16_t adc_read[ADC_CHANNELS];
static uint16_t adc_hist[ADC_CHANNELS][ADC_HISTLEN];
static uint16_t adc_avg[ADC_CHANNELS];

static uint8_t adc_idx = ADC_HISTLEN;

static int16_t vref;
static uint8_t calibrate = 0;



void adc_init()
{
    // ensure ADC is disabled
    ADC1->CR = 0;
    
    // calibration configuration
    ADC1->CCSR = LL_ADC_CAL_SAMPLINGTIME_8CYCLES | ADC_CCSR_CALSEL;

    // enable VREFINT and temperature sensor
    ADC->CCR = ADC_CCR_TSEN | ADC_CCR_VREFEN;

    // configure scan channels, sampling time
    ADC1->CHSELR = CONF_SET1_AN | PROBE_AN | 
                    CONF_MODE_AN | CONF_SET0_AN |
                    ADC_CHSELR_CHSEL11 | ADC_CHSELR_CHSEL12;
    ADC1->SMPR = SAMPLE_TIME;

    // by default, DMA selection for all channels is ADC
    // so all we need to do is set the peripheral address
    DMA1_Channel1->CPAR  = (uint32_t)&ADC1->DR;
}

void adc_next()
{
	uint8_t i, j;
	int16_t w;

    // start ADC calibration
    if (calibrate) {
		calibrate = 0;

		// stop if conversion in progress (it shouldn't ever be)
		if (ADC1->CR & ADC_CR_ADSTART) {
			ADC1->CR |= ADC_CR_ADSTP;
			while (ADC1->CR & ADC_CR_ADSTP);
		}

		// disable ADC and start calibration
        ADC1->CR &= ~(ADC_CR_ADEN);
        ADC1->CR |=   ADC_CR_ADCAL;
    }

	// copy data to history from last read
	adc_idx++;
	if (adc_idx > ADC_HISTLEN) {
		// nothing to copy when first starting
		adc_idx = 0;
		calibrate = 1;
	} else {
		// copy data to history
		for (i = 0; i < ADC_CHANNELS; i++) {
			adc_hist[i][adc_idx] = adc_read[i];
		}

		// next one
		// if we're looping, then average our last reads
		adc_idx++;
		if (adc_idx >= ADC_HISTLEN) {
			adc_idx = 0;

			// average old analog data
			for (i = 0; i < ADC_CHANNELS; i++) {
				adc_avg[i] = 0;
				for (j = 0; j < ADC_HISTLEN; j++) {
					adc_avg[i] += adc_hist[i][j];
				}
				adc_avg[i] >>= 4;
			}

			// check vref to determine if we need to recalibrate
			// the way this is written now,
			// this will always do a double-calibrate at the start
			w = adc_avg[ADC_VREF_INT];
			if ((abs(vref - w) > 3) || (abs(w - vref) > 3)) {
				calibrate = 1;
				vref = adc_avg[ADC_VREF_INT];
			}
		}
	}

	// wait for calibration to complete, if started
    while (ADC1->CR & ADC_CR_ADCAL);

    // configure and enable DMA
    DMA1_Channel1->CNDTR = ADC_CHANNELS;
    DMA1_Channel1->CMAR = (uint32_t)adc_read;
    DMA1_Channel1->CCR = LL_DMA_PRIORITY_MEDIUM | 
                    LL_DMA_MDATAALIGN_HALFWORD |
                    LL_DMA_PDATAALIGN_HALFWORD |
                    LL_DMA_MEMORY_INCREMENT |
                    LL_DMA_MODE_CIRCULAR |
                    LL_DMA_DIRECTION_PERIPH_TO_MEMORY |
                    DMA_CCR_EN;
    
    // enable and start ADC
	ADC1->ISR = 0x8e;			// clear all interrupt flags (per DS; mistranslated)
    ADC1->CR  = ADC_CR_ADEN;
    ADC1->CR |= ADC_CR_ADSTART;
}